@@ -1,10 +1,53 @@
 from multiprocessing import TimeoutError, Process, Pipe
 from multiprocessing.connection import Connection
-from typing import Any, Iterable, List, Dict, Tuple, Union
+from typing import Any, Iterable, List, Dict, Tuple, Union, Optional, Callable
 from uuid import uuid4, UUID
 import random
 import os
 from time import time
+from select import select
+from itertools import repeat
+
+# This is what we send down the pipe from parent to child
+class Request:
+    pass
+
+class Task(Request):
+    func: Callable
+    args: List
+    kwds: Dict
+    id: UUID
+
+    def __init__(self, func, args=(), kwds={}, id=None):
+        self.func = func
+        self.args = args
+        self.kwds = kwds or {}
+        self.id = id or uuid4()
+
+class QuitSignal(Request):
+    pass
+
+class RunBatchSignal(Request):
+    pass
+    
+# this is what we send back through the pipe from child to parent
+class Response:
+    id: UUID
+    pass
+
+class SuccessResponse(Response):
+    result: Any
+    def __init__(self, id: UUID, result: Any):
+        self.result = result
+        self.id = id
+
+# processing the task raised an exception
+# we save the exception in this object
+class FailResponse(Response):
+    exception: Exception
+    def __init__(self, id: UUID, exception: Exception):
+        self.id = id
+        self.exception = exception
 
 class Child:
     proc: Process
@@ -17,12 +60,12 @@ class Child:
     # does not include the termination command from parent to child
     queue_sz: int = 0
 
-    # parent_conn.send()  to give stuff to the child
+    # parent_conn.send(Request)  to give stuff to the child
     # parent_conn.recv() to get results back from child
     parent_conn: Connection
     child_conn: Connection
 
-    result_cache: Dict[UUID, Tuple[Any, Exception]] = {}
+    response_cache: Dict[UUID, Tuple[Any, Exception]] = {}
 
     _closed: bool = False
 
@@ -47,47 +90,81 @@ class Child:
     #                         {id: (None, err)} if func raised exception err
     # [None, True] -> exit gracefully (write nothing to the pipe)
     def spin(self) -> None:
+        quit_signal = False
         while True:
-            (job, quit_signal) = self.child_conn.recv()
+            req_buf = []
+            # first read in tasks until we get a pause/quit signal
+            while True:
+                request = self.child_conn.recv()
+                if isinstance(request, Task):
+                    req_buf.append(request)
+                elif isinstance(request, QuitSignal):
+                    # don't quit yet. Finish what's in the buffer.
+                    quit_signal |= True
+                    break # stop reading new stuff from the pipe
+                elif isinstance(request, RunBatchSignal):
+                    # stop reading from Pipe, process what's in the request buffer
+                    break
+            result_buf = []
+            for req in req_buf:
+                assert isinstance(req, Task)
+                # process the result
+                result = self._do_work(req)
+                result_buf.append(result)
+
+            # send all the results
+            for result in result_buf:         
+                self.child_conn.send(result)
+
             if quit_signal:
                 break
-            else:
-                (id, func, args, kwds) = job
-                result = self._do_work(id, func, args, kwds)
-                self.child_conn.send(result)
         self.child_conn.close()
 
-    def _do_work(self, id, func, args, kwds) -> Union[Tuple[Any, None], Tuple[None, Exception]]:
+    # applies the function, catching any exception if it occurs
+    def _do_work(self, task: Task) -> Response:
         try:
-            ret = {id: (func(*args, **kwds), None)}
+            result = task.func(*task.args, **task.kwds)
         except Exception as e:
             # how to handle KeyboardInterrupt?
-            ret = {id: (None, e)}
-        assert isinstance(list(ret.keys())[0], UUID)
-        return ret
+            resp = FailResponse(id=task.id, exception=e)
+        else:
+            resp = SuccessResponse(id=task.id, result=result)
+        return resp
 
-    def submit(self, func, args=(), kwds=None) -> 'AsyncResult':
+    # this sends a task to the child
+    # if as_batch=False, the child will start work on it immediately
+    # If as_batch=True, the child will load this into it's local buffer
+    # but won't start processing until we send a RunBatchSignal with self.run_batch()
+    def submit(self, func, args=(), kwds=None, as_batch=False) -> 'AsyncResult':
         if self._closed:
             raise ValueError("Cannot submit tasks after closure")
-        if kwds is None:
-            kwds = {}
-        id = uuid4()
-        self.parent_conn.send([(id, func, args, kwds), None])
+        request = Task(func=func, args=args, kwds=kwds)
+
+        self.parent_conn.send(request)
         if self.main_proc:
             self.child_conn.recv()
-            ret = self._do_work(id, func, args, kwds)
+            ret = self._do_work(request)
             self.child_conn.send(ret)
+        elif not as_batch:
+            # treat this as a batch of 1
+            self.run_batch()
         self.queue_sz += 1
-        return AsyncResult(id=id, child=self)
+        return AsyncResult(id=request.id, child=self)
+
+    # non-blocking
+    # Tells the child to start commencing work on all tasks sent up until now
+    def run_batch(self):
+        self.parent_conn.send(RunBatchSignal())
 
     # grab all results in the pipe from child to parent
-    # save them to self.result_cache
-    def flush(self):
+    # save them to self.response_cache
+    def flush_results(self):
         # watch out, when the other end is closed, a termination byte appears, so .poll() returns True
         while (not self.parent_conn.closed) and (self.queue_sz > 0) and self.parent_conn.poll(0):
             result = self.parent_conn.recv()
-            assert isinstance(list(result.keys())[0], UUID)
-            self.result_cache.update(result)
+            id = result.id
+            assert id not in self.response_cache
+            self.response_cache[id] = result
             self.queue_sz -= 1
 
     # prevent new tasks from being submitted
@@ -97,10 +174,10 @@ class Child:
         if not self._closed:
             if not self.main_proc:
                 # send quit signal to child
-                self.parent_conn.send([None, True])
+                self.parent_conn.send(QuitSignal())
             else:
                 # no child process to close
-                self.flush()
+                self.flush_results()
                 self.child_conn.close()
 
             # keep track of closure,
@@ -122,7 +199,7 @@ class Child:
             finally:
                 self.proc.close()
 
-        self.flush()
+        self.flush_results()
         self.parent_conn.close()
 
 
@@ -155,13 +232,13 @@ class AsyncResult:
         assert isinstance(id, UUID)
         self.id = id
         self.child = child
-        self.result: Union[Tuple[Any, None], Tuple[None, Exception]] = None
+        self.response: Result = None
 
-    # assume the result is in the self.child.result_cache
-    # move it into self.result
+    # assume the result is in the self.child.response_cache
+    # move it into self.response
     def _load(self):
-        self.result = self.child.result_cache[self.id]
-        del self.child.result_cache[self.id] # prevent memory leak
+        self.response = self.child.response_cache[self.id]
+        del self.child.response_cache[self.id] # prevent memory leak
 
     # Return the result when it arrives.
     # If timeout is not None and the result does not arrive within timeout seconds
@@ -169,15 +246,15 @@ class AsyncResult:
     # If the remote call raised an exception then that exception will be reraised by get().
     # .get() must remember the result
     # and return it again multiple times
-    # delete it from the Child.result_cache to avoid memory leak
+    # delete it from the Child.response_cache to avoid memory leak
     def get(self, timeout=None):
-        if self.result is not None:
-            (response, ex) = self.result
-            if ex:
-                raise ex
-            else:
-                return response
-        elif self.id in self.child.result_cache:
+        if self.response is not None:
+            if isinstance(self.response, SuccessResponse):
+                return self.response.result
+            elif isinstance(self.response, FailResponse):
+                assert isinstance(self.response.exception, Exception)
+                raise self.response.exception
+        elif self.id in self.child.response_cache:
             self._load()
             return self.get(0)
         else:
@@ -189,12 +266,12 @@ class AsyncResult:
 
     # Wait until the result is available or until timeout seconds pass.
     def wait(self, timeout=None):
-        start_t = time()
-        if self.result is None:
-            self.child.flush()
+        if self.response is None:
+            start_t = time()
+            self.child.flush_results()
             # the result we want might not be the next result
             # it might be the 2nd or 3rd next
-            while (self.id not in self.child.result_cache) and \
+            while (self.id not in self.child.response_cache) and \
                   ((timeout is None) or (time() - timeout < start_t)):
                 if timeout is None:
                     self.child.parent_conn.poll()
@@ -203,23 +280,86 @@ class AsyncResult:
                     remaining = timeout - elapsed_so_far
                     self.child.parent_conn.poll(remaining)
                 if self.child.parent_conn.poll(0):
-                    self.child.flush()
+                    self.child.flush_results()
 
     # Return whether the call has completed.
     def ready(self):
-        self.child.flush()
-        return self.result or (self.id in self.child.result_cache)
+        self.child.flush_results()
+        return self.response or (self.id in self.child.response_cache)
 
     # Return whether the call completed without raising an exception.
     # Will raise ValueError if the result is not ready.
     def successful(self):
-        if self.result is None:
+        if self.response is None:
             if not self.ready():
                 raise ValueError("Result is not ready")
             else:
                 self._load()
 
-        return self.result[1] is None
+        return isinstance(self.response, SuccessResponse)
+
+# map_async and starmap_async return a single AsyncResult
+# which is a list of actual results
+# This class aggregates many AsyncResult into one
+class AsyncResultList(AsyncResult):
+    def __init__(self, child_results: List[AsyncResult]):
+        self.child_results = child_results
+        self.result: List[Union[Tuple[Any, None], Tuple[None, Exception]]] = None
+
+    # assume the result is in the self.child.response_cache
+    # move it into self.result
+    def _load(self):
+        for c in self.child_results:
+            c._load()
+
+    # Return the result when it arrives.
+    # If timeout is not None and the result does not arrive within timeout seconds
+    # then multiprocessing.TimeoutError is raised.
+    # If the remote call raised an exception then that exception will be reraised by get().
+    # .get() must remember the result
+    # and return it again multiple times
+    # delete it from the Child.response_cache to avoid memory leak
+    def get(self, timeout=None):
+        self.wait(timeout)
+        assert self.ready()
+
+        results = []
+        for (i, c) in enumerate(self.child_results):
+            try:
+                result = c.get(0)
+            except Exception:
+                print(f"Exception raised for {i}th task out of {len(self.child_results)}")
+                for c2 in self.child_results[i+1:]:
+                    c2.child.flush_results()
+                raise
+
+            results.append(result)
+        return results
+
+    # Wait until the result is available or until timeout seconds pass.
+    def wait(self, timeout=None):
+        
+        if timeout:
+            end_t = time() + timeout
+        else:
+            end_t = None
+        for c in self.child_results:
+            # Consider cumulative timeout
+            if timeout is not None:
+                timeout_remaining = end_t - time()
+            else:
+                timeout_remaining = None
+
+            c.wait(timeout_remaining)
+
+    # Return whether the call has completed.
+    def ready(self):
+        return all(c.ready() for c in self.child_results)
+
+    # Return whether the call completed without raising an exception.
+    # Will raise ValueError if the result is not ready.
+    def successful(self):
+        return all(c.successful() for c in self.child_results)
 
 class Pool:
     def __init__(self, processes=None, initializer=None, initargs=None, maxtasksperchild=None, context=None):
@@ -245,15 +385,14 @@ class Pool:
 
         self._closed = False
 
-    def __enter__(self):
-        self._closed = False
-
         if self.num_processes > 0:
             self.children = [Child() for _ in range(self.num_processes)]
         else:
             # create one 'child' which will just do work in the main thread
             self.children = [Child(main_proc=True)]
 
+
+    def __enter__(self):
         return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
@@ -294,36 +433,106 @@ class Pool:
         if error_callback:
             raise NotImplementedError("error_callback not implemented")
 
-        if self._closed:
-            raise ValueError("Pool already closed")
-        if kwds is None:
-            kwds = {}
+        results = self._apply_batch_async(func, [args], [kwds])
+        assert len(results) == 1
+        return results[0]
 
 
-        # choose the first idle process if there is one
-        # if not, choose the process with the shortest queue
-        for c in self.children:
-            c.flush()
-        min_q_sz = min(c.queue_sz for c in self.children)
-        c = random.choice([c for c in self.children if c.queue_sz == min_q_sz])
-        return c.submit(func, args, kwds)
-
-    def map_async(self, func, iterable, chunksize=None, callback=None, error_callback=None) -> List[AsyncResult]:
+    def map_async(self, func, iterable, chunksize=None, callback=None, error_callback=None) -> AsyncResult:
         return self.starmap_async(func, zip(iterable), chunksize, callback, error_callback)
 
     def map(self, func, iterable, chunksize=None, callback=None, error_callback=None) -> List:
         return self.starmap(func, zip(iterable), chunksize, callback, error_callback)
 
-    def starmap_async(self, func, iterable: Iterable[Iterable], chunksize=None, callback=None, error_callback=None) -> List[AsyncResult]:
+    def starmap_async(self, func, iterable: Iterable[Iterable], chunksize=None, callback=None, error_callback=None) -> AsyncResult:
         if chunksize:
             raise NotImplementedError("Haven't implemented chunksizes. Infinite chunksize only.")
         if callback or error_callback:
             raise NotImplementedError("Haven't implemented callbacks")
-        return [self.apply_async(func, args) for args in iterable]
+
+        results = self._apply_batch_async(func, args_iterable=iterable, kwds_iterable=repeat({}))
+        result = AsyncResultList(child_results=results)
+        return result
+
+    # like starmap, but has argument for keyword args
+    # so apply_async can call this
+    # (apply_async supports kwargs but starmap_async does not)
+    def _apply_batch_async(self, func, args_iterable: Iterable[Iterable], kwds_iterable: Iterable[Dict]) -> List[AsyncResult]:
+        if self._closed:
+            raise ValueError("Pool already closed")
+
+        for c in self.children:
+            c.flush_results()
+
+        results = []
+        children_called = set()
+        for (args, kwds) in zip(args_iterable, kwds_iterable):
+            child = self._choose_child() # already flushed results
+            children_called.add(child)
+            result = child.submit(func, args, kwds, as_batch=True)
+            results.append(result)
+
+        for child in children_called:
+            child.run_batch()
+
+        return results
+        
+
+    # return the child with the shortest queue
+    # if a tie, choose randomly
+    # You should call c.flush_results() first before calling this
+    def _choose_child(self) -> Child:
+        min_q_sz = min(c.queue_sz for c in self.children)
+        return random.choice([c for c in self.children if c.queue_sz <= min_q_sz])
+
 
     def starmap(self, func, iterable: Iterable[Iterable], chunksize=None, callback=None, error_callback=None) -> List[Any]:
-        results = self.starmap_async(func, iterable, chunksize, callback, error_callback)
-        return [r.get() for r in results]
+        if chunksize:
+            raise NotImplementedError("chunksize not implemented")
+        if callback:
+            raise NotImplementedError("callback not implemented")
+        if error_callback:
+            raise NotImplementedError("error_callback not implemented")
+
+        idle_children = set(self.children)
+        ids = []
+        pending_results: Dict[UUID, AsyncResult] = {}
+
+        for args in iterable:
+            if not idle_children:
+                # wait for a child to become idle
+                # by waiting for any of the pipes from children to become readable
+                ready, _, _ = select([c.parent_conn for c in self.children], [], [])
+
+                # at least one child is idle. 
+                # check all children, read their last result from the pipe
+                # then issue the new task
+                for child in self.children:
+                    if child.parent_conn in ready:
+                        assert child.parent_conn.poll()
+                        child.flush_results()
+                        idle_children.add(child)
+                
+
+            child = idle_children.pop()
+            async_result = child.submit(func, args)
+            pending_results[async_result.id] = async_result
+            ids.append(async_result.id)
+
+
+        if len(idle_children) < len(self.children):
+            # if at least one child is still working
+            # wait with select
+            ready, _, _ = select([c.parent_conn for c in self.children if c not in idle_children], [], [])
+
+        # get all the results
+        # re-arranging the order
+        results = []
+        for (i, id) in enumerate(ids):
+            result = pending_results[id].get()
+            results.append(result)
+
+        return results
 
     def imap(self, func, iterable, chunksize=None):
         raise NotImplementedError("Only normal apply, map, starmap and their async equivilents have been implemented")
