@@ -1,18 +1,23 @@
 import unittest
-import multiprocessing
+import multiprocessing, multiprocessing.pool
 from lambda_multiprocessing import Pool, TimeoutError, AsyncResult
 from time import time, sleep
-from typing import Tuple
+from typing import Tuple, Optional
 from pathlib import Path
 import os
+import sys
+from functools import cache
 
 import boto3
-from moto import mock_s3
+from moto import mock_aws
+from lambda_multiprocessing.timeout import TimeoutManager, TestTimeoutException
 
 # add an overhead for duration when asserting the duration of child processes
 # if other processes are hogging CPU, make this bigger
 delta = 0.1
 
+SEC_PER_MIN = 60
+
 # some simple functions to run inside the child process
 def square(x):
     return x*x
@@ -29,6 +34,19 @@ def divide(a, b):
 def return_args_kwargs(*args, **kwargs):
     return {'args': args, 'kwargs': kwargs}
 
+def return_with_sleep(x, delay=0.3):
+    sleep(delay)
+    return x
+
+def _raise(ex: Optional[Exception]):
+    if ex:
+        raise ex
+
+class ExceptionA(Exception):
+    pass
+class ExceptionB(Exception):
+    pass
+
 class TestStdLib(unittest.TestCase):
     @unittest.skip('Need to set up to remove /dev/shm')
     def test_standard_library(self):
@@ -38,13 +56,25 @@ class TestStdLib(unittest.TestCase):
 
 # add assertDuration
 class TestCase(unittest.TestCase):
+
+    max_timeout = SEC_PER_MIN*2
+    timeout_mgr = TimeoutManager(seconds=max_timeout)
+
+    def setUp(self):
+        self.timeout_mgr.start()
+
+    def tearDown(self):
+        self.timeout_mgr.stop()
+
     # use like
     # with self.assertDuration(1, 2):
     #   something
     # to assert that something takes between 1 to 2 seconds to run
+    # If the task takes forever, this assertion will not be raised
+    # For a potential eternal task, use timeout.TimeoutManager
     def assertDuration(self, min_t=None, max_t=None):
         class AssertDuration:
-            def __init__(self, test):
+            def __init__(self, test: unittest.TestCase):
                 self.test = test
             def __enter__(self):
                 self.start_t = time()
@@ -57,10 +87,12 @@ class TestCase(unittest.TestCase):
                         self.test.assertGreaterEqual(duration, min_t, f"Took less than {min_t}s to run")
                     if max_t is not None:
                         self.test.assertLessEqual(duration, max_t, f"Took more than {max_t}s to run")
+                return False
 
         return AssertDuration(self)
 
 
+
 class TestAssertDuration(TestCase):
 
     def test(self):
@@ -78,22 +110,23 @@ class TestAssertDuration(TestCase):
             sleep(t)
 
 class TestApply(TestCase):
+    pool_generator = Pool
 
     def test_simple(self):
         args = range(5)
-        with Pool() as p:
+        with self.pool_generator() as p:
             actual = [p.apply(square, (x,)) for x in args]
         with multiprocessing.Pool() as p:
             expected = [p.apply(square, (x,)) for x in args]
         self.assertEqual(actual, expected)
 
     def test_two_args(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             actual = p.apply(sum_two, (3, 4))
             self.assertEqual(actual, sum_two(3, 4))
 
     def test_kwargs(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             actual = p.apply(return_args_kwargs, (1, 2), {'x': 'X', 'y': 'Y'})
             expected = {
                 'args': (1,2),
@@ -104,10 +137,16 @@ class TestApply(TestCase):
             }
             self.assertEqual(actual, expected)
 
+# rerun all the tests with stdlib
+# to confirm we have the same behavior
+class TestApplyStdLib(TestApply):
+    pool_generator = multiprocessing.Pool
+
 class TestApplyAsync(TestCase):
+    pool_generator = Pool
 
     def test_result(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             r = p.apply_async(square, (2,))
             result = r.get(2)
             self.assertEqual(result, square(2))
@@ -115,7 +154,7 @@ class TestApplyAsync(TestCase):
     def test_twice(self):
         args = range(5)
         # now try twice with the same pool
-        with Pool() as p:
+        with self.pool_generator() as p:
             ret = [p.apply_async(square, (x,)) for x in args]
             ret.extend(p.apply_async(square, (x,)) for x in args)
 
@@ -124,7 +163,7 @@ class TestApplyAsync(TestCase):
     def test_time(self):
         # test the timing to confirm it's in parallel
         n = 4
-        with Pool(n) as p:
+        with self.pool_generator(n) as p:
             with self.assertDuration(min_t=n-1, max_t=(n-1)+delta):
                 with self.assertDuration(max_t=1):
                     ret = [p.apply_async(sleep, (r,)) for r in range(n)]
@@ -135,14 +174,14 @@ class TestApplyAsync(TestCase):
         # .get after __exit__, but process finishes before __exit__
         with self.assertRaises(multiprocessing.TimeoutError):
             t = 1
-            with Pool() as p:
+            with self.pool_generator() as p:
                 r = p.apply_async(square, (t,))
                 sleep(1)
             r.get()
 
         with self.assertRaises(multiprocessing.TimeoutError):
             t = 1
-            with Pool() as p:
+            with self.pool_generator() as p:
                 r = p.apply_async(square, (t,))
                 sleep(1)
             r.get(1)
@@ -150,26 +189,26 @@ class TestApplyAsync(TestCase):
         # __exit__ before process finishes
         with self.assertRaises(multiprocessing.TimeoutError):
             t = 1
-            with Pool() as p:
+            with self.pool_generator() as p:
                 r = p.apply_async(sleep, (t,))
             r.get(t+1) # .get() with arg after __exit__
 
         with self.assertRaises(multiprocessing.TimeoutError):
             t = 1
-            with Pool() as p:
+            with self.pool_generator() as p:
                 r = p.apply_async(sleep, (t,))
             r.get() # .get() without arg after __exit__
 
     def test_get_not_ready_a(self):
         t = 2
-        with Pool() as p:
+        with self.pool_generator() as p:
             r = p.apply_async(sleep, (t,))
             with self.assertRaises(multiprocessing.TimeoutError):
                 r.get(t-1) # result not ready get
 
     def test_get_not_ready_b(self):
         t = 2
-        with Pool() as p:
+        with self.pool_generator() as p:
             # check same exception exists from main
             r = p.apply_async(sleep, (t,))
             with self.assertRaises(TimeoutError):
@@ -177,7 +216,7 @@ class TestApplyAsync(TestCase):
 
     def test_get_not_ready_c(self):
         t = 2
-        with Pool() as p:
+        with self.pool_generator() as p:
             r = p.apply_async(sleep, (t,))
             sleep(1)
             self.assertFalse(r.ready())
@@ -185,7 +224,7 @@ class TestApplyAsync(TestCase):
             self.assertTrue(r.ready())
 
     def test_wait(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             r = p.apply_async(square, (1,))
             with self.assertDuration(max_t=delta):
                 r.wait()
@@ -209,13 +248,13 @@ class TestApplyAsync(TestCase):
             ret = r.get(0)
 
     def test_get_twice(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             r = p.apply_async(square, (2,))
             self.assertEqual(r.get(), square(2))
             self.assertEqual(r.get(), square(2))
 
     def test_successful(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             r = p.apply_async(square, (1,))
             sleep(delta)
             self.assertTrue(r.successful())
@@ -233,13 +272,13 @@ class TestApplyAsync(TestCase):
             self.assertFalse(r.successful())
 
     def test_two_args(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             ret = p.apply_async(sum_two, (1, 2))
             ret.wait()
             self.assertEqual(ret.get(), sum_two(1,2))
 
     def test_kwargs(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             actual = p.apply_async(return_args_kwargs, (1, 2), {'x': 'X', 'y': 'Y'}).get()
             expected = {
                 'args': (1,2),
@@ -252,14 +291,21 @@ class TestApplyAsync(TestCase):
 
     def test_error_handling(self):
         with self.assertRaises(AssertionError):
-            with Pool() as p:
+            with self.pool_generator() as p:
                 r = p.apply_async(fail, (1,))
                 r.get()
 
+# retrun tests with standard library
+# to confirm our behavior matches theirs
+class TestApplyAsyncStdLib(TestApplyAsync):
+    pool_generator = multiprocessing.Pool
+
 class TestMap(TestCase):
+    pool_generator = Pool
+    
     def test_simple(self):
         args = range(5)
-        with Pool() as p:
+        with self.pool_generator() as p:
             actual = p.map(square, args)
         self.assertIsInstance(actual, list)
         expected = [square(x) for x in args]
@@ -268,93 +314,136 @@ class TestMap(TestCase):
 
     def test_duration(self):
         n = 2
-        with Pool(n) as p:
+        with self.pool_generator(n) as p:
             with self.assertDuration(min_t=(n-1)-delta, max_t=(n+1)+delta):
                 p.map(sleep, range(n))
 
     def test_error_handling(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             with self.assertRaises(AssertionError):
                 p.map(fail, range(2))
 
     @unittest.skip('Need to implement chunking to fix this')
     def test_long_iter(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             p.map(square, range(10**3))
 
-class TestMapAsync(TestCase):
-    def test_simple(self):
-        args = range(5)
-        with Pool() as p:
-            actual = p.map_async(square, args)
-            self.assertIsInstance(actual, list)
-            for x in actual:
-                self.assertIsInstance(x, AsyncResult)
+    def test_without_with(self):
+        p = self.pool_generator(3)
+        ret = p.map(square, [1,2])
+        self.assertEqual(ret, [1,2*2])
+        p.close()
+        p.join()
 
-            results = [a.get() for a in actual]
-        self.assertEqual(results, [square(e) for e in args])
+class TestMapStdLib(TestMap):
+    pool_generator = multiprocessing.Pool
+
+class TestMapAsync(TestCase):
+    pool_generator = Pool
+
+    def test_simple(self):
+        num_payloads = 5
+        args = range(num_payloads)
+        for num_procs in [(num_payloads+1), (num_payloads-1)]:
+            with self.pool_generator(num_procs) as p:
+                actual = p.map_async(square, args)
+                self.assertIsInstance(actual, (AsyncResult, multiprocessing.pool.AsyncResult))
+                results = actual.get()
+            self.assertEqual(results, [square(e) for e in args])
 
     def test_duration(self):
-        n = 2
-        with Pool(n) as p:
-            with self.assertDuration(min_t=(n-1)-delta, max_t=(n+1)+delta):
+        sleep_duration = 0.5
+        n_procs = 2
+        num_tasks_per_proc = 3
+        expected_wall_time = sleep_duration * num_tasks_per_proc
+        with self.pool_generator(n_procs) as p:
+            with self.assertDuration(min_t=expected_wall_time-delta, max_t=expected_wall_time+delta):
                 with self.assertDuration(max_t=delta):
-                    results = p.map_async(sleep, range(n))
-                [r.get() for r in results]
+                    results = p.map_async(sleep, [sleep_duration] * (num_tasks_per_proc * n_procs))
+                results.get()
+       
 
     def test_error_handling(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             r = p.map_async(fail, range(2))
             with self.assertRaises(AssertionError):
-                [x.get(1) for x in r]
+                r.get()
+
+    def test_multi_error(self):
+        # standard library can raise either error
+
+        with self.assertRaises((ExceptionA, ExceptionB)):
+            with self.pool_generator() as p:
+                r = p.map_async(_raise, (None, ExceptionA("Task 1"), None, ExceptionB("Task 3")))
+                r.get()
+
+class TestMapAsyncStdLib(TestMapAsync):
+    pool_generator = multiprocessing.Pool
 
 class TestStarmap(TestCase):
+    pool_generator = Pool
+
     def test(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             actual = p.starmap(sum_two, [(1,2), (3,4)])
         expected = [(1+2), (3+4)]
         self.assertEqual(actual, expected)
 
     def test_error_handling(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             with self.assertRaises(ZeroDivisionError):
                 p.starmap(divide, [(1,2), (3,0)])
 
-
+class TestStarmapStdLib(TestStarmap):
+    pool_generator = multiprocessing.Pool
 
 class TestStarmapAsync(TestCase):
+    pool_generator = Pool
+
     def test(self):
-        with Pool() as p:
-            actual = p.starmap_async(sum_two, [(1,2), (3,4)])
-            self.assertIsInstance(actual, list)
-            actual = [r.get() for r in actual]
+        with self.pool_generator() as p:
+            response = p.starmap_async(sum_two, [(1,2), (3,4)])
+            self.assertIsInstance(response, (AsyncResult, multiprocessing.pool.AsyncResult))
+            actual = response.get()
         expected = [(1+2), (3+4)]
         self.assertEqual(actual, expected)
 
     def test_error_handling(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             results = p.starmap_async(divide, [(1,2), (3,0)])
             with self.assertRaises(ZeroDivisionError):
-                [r.get() for r in results]
+                results.get()
 
 
-class TestTidyUp(TestCase):
+
+class TestStarmapAsyncStdLib(TestStarmapAsync):
+    pool_generator = multiprocessing.Pool
+
+class TestExit(TestCase):
+    # only test this with our library,
+    # not the standard library
+    # because the standard library has a bug
+    # https://github.com/python/cpython/issues/79659
+
     # test that the implicit __exit__
     # waits for child process to finish
     def test_exit(self):
         t = 1
         with Pool() as p:
-            r = p.apply_async(sleep, (t,))
             t1 = time()
+            r = p.apply_async(sleep, (t,))
         t2 = time()
         self.assertLessEqual(abs((t2-t1)-t), delta)
 
+class TestTidyUp(TestCase):
+    pool_generator = Pool
+
     # test that .close() stops new submisssions
     # but does not halt existing num_processes
     # nor wait for them to finish
     def test_close(self):
         t = 1
-        with Pool() as p:
+        with self.pool_generator() as p:
             r = p.apply_async(sleep, (t,))
             with self.assertDuration(max_t=delta):
                 p.close()
@@ -363,7 +452,7 @@ class TestTidyUp(TestCase):
             pass # makes traceback from __exit__ clearer
 
     def test_submit_after_close(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             p.close()
             with self.assertRaises(ValueError):
                 p.apply_async(square, (1,))
@@ -372,7 +461,7 @@ class TestTidyUp(TestCase):
     # wait for child process to finish
     def test_terminate(self):
         with self.assertDuration(max_t=delta):
-            with Pool() as p:
+            with self.pool_generator() as p:
                 r = p.apply_async(sleep, (1,))
                 t1 = time()
                 p.terminate()
@@ -380,11 +469,102 @@ class TestTidyUp(TestCase):
                 self.assertLessEqual(t2-t1, delta)
 
     def test_submit_after_terminate(self):
-        with Pool() as p:
+        with self.pool_generator() as p:
             p.terminate()
             with self.assertRaises(ValueError):
                 p.apply_async(square, (1,))
 
+class TestTidyUpStdLib(TestTidyUp):
+    pool_generator = multiprocessing.Pool
+
+class TestDeadlock(TestCase):
+
+    # test this issue:
+    # https://github.com/mdavis-xyz/lambda_multiprocessing/issues/17
+    def test_map_deadlock(self):
+
+        child_sleep = 0.01
+        num_payloads = 6
+
+        # use standard library to start with
+        # and to measure a 'normal' duration
+        # (the time spend passing data between processes is longer than the sleep
+        #  inside the worker)
+        expected_duration = child_sleep * num_payloads
+        data = [self.generate_big_data() for _ in range(num_payloads)]
+        start_t = time()
+
+        with multiprocessing.Pool(processes=1) as p:
+            p.map(return_with_sleep, data)
+        
+        end_t = time()
+        stdlib_duration = end_t - start_t
+
+        # this timeout manager doesn't work
+        # need to run the parent inside another process/thread?
+        # now our one
+        data = [self.generate_big_data() for _ in range(num_payloads)]
+        with Pool(processes=1) as p:
+            with TimeoutManager(stdlib_duration*2, "This Library's map deadlocked"):
+                try:
+                    p.map(return_with_sleep, data)
+                except TestTimeoutException:
+                    p.terminate()
+                    raise
+
+    def test_map_async_deadlock(self):
+
+        child_sleep = 0.4
+        num_payloads = 3
+        data = [self.generate_big_data() for _ in range(num_payloads)]
+
+        # use standard library to start with
+        # and to measure a 'normal' duration
+        # (the time spend passing data between processes is longer than the sleep
+        #  inside the worker)
+        expected_duration = child_sleep * num_payloads
+        start_t = time()
+        with multiprocessing.Pool(processes=1) as p:
+            results = p.map_async(return_with_sleep, data)
+            results.get()
+        end_t = time()
+        stdlib_duration = end_t - start_t
+
+        # now our one
+        with Pool(processes=1) as p:
+            with TimeoutManager(stdlib_duration*2, "This Library's map_async deadlocked"):
+                try:
+                    results = p.map_async(return_with_sleep, data)
+                    results.get()
+                except TestTimeoutException:
+                    p.terminate()
+                    raise
+                
+    # test that map_async returns immediately
+    # even when there are multiple tasks per child
+    # with payloads bigger than the buffer
+    def test_nonblocking(self):
+        sleep_duration = 2
+        n_procs = 2
+        num_tasks_per_proc = 3
+        expected_wall_time = sleep_duration * num_tasks_per_proc
+
+        args = [(self.generate_big_data(), sleep_duration)] * (num_tasks_per_proc * n_procs)
+        with Pool(n_procs) as p:
+            try:
+                with TimeoutManager(sleep_duration * num_tasks_per_proc * 2, "This Library's map_async deadlocked"):
+                    with self.assertDuration(max_t=sleep_duration*0.5):
+                        results = p.map_async(return_with_sleep, args)
+                results.get(expected_wall_time*1.5)
+            except Exception:
+                p.terminate()
+                raise
+
+    @classmethod
+    @cache
+    def generate_big_data(cls, sz=2**24) -> bytes:
+        return 'x' * sz
+
 
 # must be a global method to be pickleable
 def upload(args: Tuple[str, str, bytes]):
@@ -401,16 +581,17 @@ class TestMoto(TestCase):
                 r = p.apply_async(sleep, (t,))
                 pass # makes traceback from __exit__ easier to read
 
-    @mock_s3
+    @mock_aws
     def test_moto(self):
         bucket_name = 'mybucket'
         key = 'my-file'
         data = b"123"
-        client = boto3.client('s3')
+        region = os.getenv("AWS_DEFAULT_REGION", "ap-southeast-2")
+        client = boto3.client('s3', region_name=region)
         client.create_bucket(
             Bucket=bucket_name,
             CreateBucketConfiguration={
-                'LocationConstraint': 'ap-southeast-2'
+                'LocationConstraint': region
             },
         )
         # upload in a different thread
